/*
    *
    * This file is a part of CoreShot.
    * A screen capture utility for CuboCore Application Suite
    * Copyright 2019 CuboCore Group
    *

    *
    * This program is free software; you can redistribute it and/or modify
    * it under the terms of the GNU General Public License as published by
    * the Free Software Foundation; either version 3 of the License, or
    * (at your option) any later version.
    *

    *
    * This program is distributed in the hope that it will be useful,
    * but WITHOUT ANY WARRANTY; without even the implied warranty of
    * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    * GNU General Public License for more details.
    *

    *
    * You should have received a copy of the GNU General Public License
    * along with this program; if not, write to the Free Software
    * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
    * MA 02110-1301, USA.
    *
*/


#include <QMessageBox>
#include <QDesktopWidget>
#include <QPushButton>
#include <QTimer>
#include <QScreen>
#include <QPalette>
#include <QDebug>
#include <QX11Info>
#include <QDateTime>
#include <QMouseEvent>
#include <X11/Xlib.h>
#include <X11/Xatom.h>

#include <cprime/cprime.h>

#include "previewwidget.h"
#include "coreshot.h"
#include "coreshotdialog.h"
#include "ui_coreshotdialog.h"


coreshotdialog::coreshotdialog(QWidget *parent) : QWidget(parent), ui(new Ui::coreshotdialog)
{
    ui->setupUi(this);

    mousegrabbed = false;
    areaOverlay = nullptr;
    startsetup();

    // Set window as a framless
    this->setWindowFlags(Qt::Window | Qt::FramelessWindowHint);

    // Set the window size & position
    QScreen *scr = QGuiApplication::primaryScreen();
    int mx = (scr->availableGeometry().width() / 2) - ((this->width() / 2) + 10);
    int my = (scr->availableGeometry().height()) - (this->height() + 15);
    this->move(mx, my);
}

coreshotdialog::~coreshotdialog()
{
    delete ui;
}

void coreshotdialog::startsetup()
{
    connect(ui->captureScreen, &QPushButton::clicked, [this]() {
        this->setWindowState(Qt::WindowMinimized);

        if (windowState() == Qt::WindowMinimized) {
            QTimer::singleShot((ui->delay->value() * 1000) + 200, this, SLOT(shootFullScreen()));
        }
    });

    connect(ui->captureWindow, &QPushButton::clicked, [this]() {
        this->setWindowState(Qt::WindowMinimized);

        if (windowState() == Qt::WindowMinimized) {
            QTimer::singleShot((ui->delay->value() * 1000) + 200, this, SLOT(shootActiveWindow()));
        }
    });

    connect(ui->captureSelection, &QPushButton::clicked, this, &coreshotdialog::shootSelectArea);

    connect(ui->cancel, &QPushButton::clicked, this, &coreshotdialog::close);
}

void coreshotdialog::passToShotPreview()
{
    bool autosave = sm->value("CoreShot", "AutoSave");
    QString saveLocation = QString(sm->value("CoreShot", "SaveLocation"));

    if (autosave) {
        QString dirName = saveLocation;
        dirName = CPrime::ValidityFunc::checkIsValidDir(dirName);

        if (dirName.count()) {
            QFile file(dirName + "/Screenshot_" + QDateTime::currentDateTime().toString("yyyy-MM-dd_hh-mm-ss") + ".png");
            m_pixmap.save(&file, "PNG");
            file.close();

            this->close();

            // Function from utilities.cpp
            CPrime::InfoFunc::messageEngine("Screenshot Saved", CPrime::MessageType::Info, this);
            return;
        } else {
            CPrime::InfoFunc::messageEngine("Screenshot Not Saved\nPlease select a valid directory from CoreGarage.", CPrime::MessageType::Warning, this);
        }

        // Based on Message engine time how it will stay
        QTimer::singleShot(1000, this, SLOT(close()));
    } else {
        coreshot *shootP = new coreshot();
        shootP->setPixmap(m_pixmap);
        shootP->show();
        this->close();
    }
}

void coreshotdialog::shootFullScreen()
{
    QScreen *scrn = QApplication::screens().at(0);
    m_pixmap = scrn->grabWindow(QApplication::desktop()->winId());
    passToShotPreview();
}

void coreshotdialog::shootSelectArea()
{
    if (mousegrabbed) {
        return;
    }

    lastgeom = this->geometry();

    snapArea = QRect();
    this->grabMouse(QCursor(Qt::CrossCursor));
    mousegrabbed = true;
    this->ui->frame->setEnabled(false);
    this->setWindowOpacity(0);
}

void coreshotdialog::shootActiveWindow()
{
    // Current window rectangle.
    QRect rect = getWindowFrame(getActiveWindowId());

    // Current whole screen.
    QScreen *screen = QGuiApplication::primaryScreen();

    auto cwposX = rect.topLeft().x();
    auto cwposY = rect.topLeft().y();
    auto cwHeight = rect.topLeft().y() + rect.height();
    auto cwWidth = rect.topLeft().x() + rect.width();
    auto scHeight = screen->size().height();
    auto scWidth = screen->size().width();
    auto scposX = screen->availableGeometry().x();
    auto scposY = screen->availableGeometry().y();

    if (cwHeight > scHeight) {
        cwHeight = rect.height() - (cwHeight - screen->availableSize().height());

        if (cwposY < 0) {
            cwposY = scposY;
            cwHeight = cwHeight + rect.y();
        } else {
            cwHeight = cwHeight + scposY;
        }
    } else {
        cwHeight = rect.height();

        if (cwposY < 0) {
            cwposY = scposY;
            cwHeight = rect.y() + rect.height() - cwposY;
        }
    }

    if (cwWidth > scWidth) {
        cwWidth = rect.width() - (cwWidth - screen->availableSize().width());

        if (cwposX < 0) {
            cwposX = scposX;
            cwWidth = cwWidth + rect.x();
        } else {
            cwWidth = cwWidth + scposX;
        }
    } else {
        cwWidth = rect.width();

        if (cwposX < 0) {
            cwposX = scposX;
            cwWidth = rect.x() + rect.width() - cwposX;
        }
    }

    auto pixmap = screen->grabWindow(QApplication::desktop()->winId(),
                                     cwposX, cwposY, cwWidth, cwHeight);

    m_pixmap = pixmap;

    this->hide();
    passToShotPreview();
}

WId coreshotdialog::getActiveWindowId()
{
    WId root = WId(QX11Info::appRootWindow());
    Atom atom = XInternAtom(QX11Info::display(), "_NET_ACTIVE_WINDOW", false);
    unsigned long type, resultLen, rest;
    int format;
    WId result = 0;
    unsigned char *data = nullptr;

    // FIX ME
    // Don't know the new cast for this 'XA_CARDINAL' because it's a Macro
    if (XGetWindowProperty(QX11Info::display(), root, atom, 0, 1, false,
                           XA_WINDOW, &type, &format, &resultLen, &rest, &data) == Success) { //cast
        result = static_cast<WId>(*reinterpret_cast<long *>(data));
        XFree(data);
    }

    return result;
}

QRect coreshotdialog::getWindowFrame(WId wid)
{
    QRect result;
    XWindowAttributes wa;

    if (XGetWindowAttributes(QX11Info::display(), wid, &wa)) {
        Window child;
        int x, y;
        // translate to root coordinate
        XTranslateCoordinates(QX11Info::display(), wid, wa.root, 0, 0, &x, &y, &child);
        result.setRect(x, y, wa.width, wa.height);

        // get the frame widths added by the window manager
        Atom atom = XInternAtom(QX11Info::display(), "_NET_FRAME_EXTENTS", false);
        unsigned long type, resultLen, rest;
        int format;
        unsigned char *data = nullptr;

        // FIX ME
        // Don't know the new cast for this 'XA_CARDINAL' because it's a Macro
        if (XGetWindowProperty(QX11Info::display(), wid, atom, 0, LONG_MAX, false,
                               (XA_CARDINAL), &type, &format, &resultLen, &rest, &data) == Success) { //cast
        }

        if (data) { // left, right, top, bottom
            long *offsets = reinterpret_cast<long *>(data);
            result.setLeft(static_cast<int>(result.left() - offsets[0]));
            result.setRight(static_cast<int>(result.right() + offsets[1]));
            result.setTop(static_cast<int>(result.top() - offsets[2]));
            result.setBottom(static_cast<int>(result.bottom() + offsets[3]));
            XFree(data);
        }
    }

    return result;
}

void coreshotdialog::mousePressEvent(QMouseEvent *ev)
{
    if (mousegrabbed) {
        pt_click = ev->globalPos();

        if (areaOverlay == nullptr) {
            areaOverlay =  new QWidget(nullptr, Qt::Window | Qt::BypassWindowManagerHint | Qt::WindowStaysOnTopHint);
            areaOverlay->setWindowOpacity(0.5);
            areaOverlay->setStyleSheet("background-color: rgba(150,150,150,120)");
        }
    }

    QWidget::mouseMoveEvent(ev);
}

void coreshotdialog::mouseMoveEvent(QMouseEvent *ev)
{
    if (mousegrabbed) {
        //Not used yet - do something to paint the area so the user can see which area is selected
        QRect area = pointsToRect(pt_click, ev->globalPos());
        areaOverlay->setGeometry(area);
        areaOverlay->show();
    } else {
        QWidget::mouseMoveEvent(ev);
    }
}

void coreshotdialog::mouseReleaseEvent(QMouseEvent *ev)
{
    if (mousegrabbed) {
        mousegrabbed = false;
        this->ui->frame->setEnabled(true);
        this->releaseMouse();
        this->setWindowOpacity(1);
        areaOverlay->hide();
        snapArea = pointsToRect(pt_click, ev->globalPos());
        this->hide();
        QTimer::singleShot(50, this, SLOT(getPix()));
    } else {
        QWidget::mouseReleaseEvent(ev); //normal processing
    }
}
QRect coreshotdialog::pointsToRect(QPoint pt1, QPoint pt2)
{
    QRect rec;

    if (pt1.x() < pt2.x()) {
        rec.setLeft(pt1.x());
        rec.setRight(pt2.x());
    } else {
        rec.setLeft(pt2.x());
        rec.setRight(pt1.x());
    }

    if (pt1.y() < pt2.y()) {
        rec.setTop(pt1.y());
        rec.setBottom(pt2.y());
    } else {
        rec.setTop(pt2.y());
        rec.setBottom(pt1.y());
    }

    return rec;
}

void coreshotdialog::getPix()
{
    QScreen *scrn = QApplication::screens().at(0);
    m_pixmap = scrn->grabWindow(QApplication::desktop()->winId(), snapArea.x(), snapArea.y(), snapArea.width(), snapArea.height());

    this->showNormal();
    this->setGeometry(lastgeom);
    passToShotPreview();
}
